import inspect
from collections.abc import Callable
from typing import ParamSpec, TypeVar

Tcov = TypeVar("Tcov", covariant=True)
TOwner = TypeVar("TOwner")
TResult = TypeVar("TResult")
TParams = ParamSpec("TParams")


def replace_self_signature(
    func: Callable[TParams, TResult],
    self_param: inspect.Parameter,
) -> Callable[TParams, TResult]:
    """
    Replaces the signature of the `self` argument of `func` with the given one.

    Arguments:
        func: The function whose self argument should be replaced.
        self_param: The new parameter description for the `self` argument.

    Returns:
        The received function with the updated annotations.

    Raises:
        ValueError: If `func` has no `self` argument.
    """
    signature = inspect.signature(func)
    if "self" not in signature.parameters:
        raise ValueError("Method has no self argument.")

    func.__signature__ = signature.replace(  # type: ignore[attr-defined]
        parameters=(
            self_param,
            *(v for k, v in signature.parameters.items() if k != "self"),
        )
    )
    return func



import asyncio
import inspect
from collections.abc import AsyncGenerator, Callable, Generator
from functools import wraps
from typing import Annotated, Any, Concatenate, Coroutine, Generic, overload

from fastapi import Depends

from .idmemo import IdMemo
from .typing import BoundMethod, Dependency, TOwner, TParams, TResult
from .utils import replace_self_signature


class SelfWrapper:
    """
    `SelfDependent` function wrappers that populate the `self` argument of the wrapped function
    from the keyword arguments (if one was provided), or use the owner instance if one wasn't
    provided. Thus the wrapped "static" methods can behave as if they were instance methods.
    """

    @classmethod
    def sync_method(
        cls,
        func: Callable[Concatenate[TOwner, TParams], TResult],
        owner: TOwner | None,
    ) -> Callable[TParams, TResult]:
        """
        Wrapper for synchronous methods.

        Arguments:
            func: The function to wrap.
            owner: An optional owner object.

        Returns:
            The wrapper.
        """

        @wraps(func)
        def do(*args: TParams.args, **kwargs: TParams.kwargs) -> TResult:
            func_self: TOwner = kwargs.pop("self") if "self" in kwargs else owner  # type: ignore[assignment]
            if func_self is None:
                raise RuntimeError("Missing self argument.")

            return func(func_self, *args, **kwargs)

        return do

    @classmethod
    def sync_generator(
        cls,
        func: Callable[Concatenate[TOwner, TParams], Generator[TResult, Any, Any]],
        owner: TOwner | None,
    ) -> Callable[TParams, Generator[TResult, None, None]]:
        """
        Wrapper for synchronous generator methods.

        Arguments:
            func: The function to wrap.
            owner: An optional owner object.

        Returns:
            The wrapper.
        """

        @wraps(func)
        def do(*args: TParams.args, **kwargs: TParams.kwargs) -> Generator[TResult, None, None]:
            func_self: TOwner = kwargs.pop("self") if "self" in kwargs else owner  # type: ignore[assignment]
            if func_self is None:
                raise RuntimeError("Missing self argument.")

            yield from func(func_self, *args, **kwargs)

        return do

    @classmethod
    def async_method(
        cls,
        func: Callable[Concatenate[TOwner, TParams], Coroutine[None, None, TResult]],
        owner: TOwner | None,
    ) -> Callable[TParams, Coroutine[None, None, TResult]]:
        """
        Wrapper for asynchronous methods.

        Arguments:
            func: The function to wrap.
            owner: An optional owner object.

        Returns:
            The wrapper.
        """

        @wraps(func)
        async def do(*args: TParams.args, **kwargs: TParams.kwargs) -> TResult:
            func_self: TOwner = kwargs.pop("self") if "self" in kwargs else owner  # type: ignore[assignment]
            if func_self is None:
                raise RuntimeError("Missing self argument.")

            return await func(func_self, *args, **kwargs)

        return do

    @classmethod
    def async_generator(
        cls,
        func: Callable[
            Concatenate[TOwner, TParams],
            AsyncGenerator[TResult, None],
        ],
        owner: TOwner | None,
    ) -> Callable[TParams, AsyncGenerator[TResult, None]]:
        """
        Wrapper for asynchronous generator methods.

        Arguments:
            func: The function to wrap.
            owner: An optional owner object.

        Returns:
            The wrapper.
        """

        @wraps(func)
        async def do(*args: TParams.args, **kwargs: TParams.kwargs) -> AsyncGenerator[TResult, None]:
            func_self: TOwner = kwargs.pop("self") if "self" in kwargs else owner  # type: ignore[assignment]
            if func_self is None:
                raise RuntimeError("Missing self argument.")

            async for res in func(func_self, *args, **kwargs):
                yield res

        return do


class SelfDependent(Generic[TOwner, TParams, TResult]):
    """
    Descriptor whose value is a method (FastAPI dependency) with an annotated
    `self` argument that can be processed and used by FastAPI as a dependency.
    """

    __slots__ = ("_wrapped", "_factory", "_memo")

    def __init__(
        self,
        wrapped: Callable[Concatenate[TOwner, TParams], TResult],
        *,
        factory: Dependency[TOwner] | None = None,
    ) -> None:
        """
        Initialization.

        Arguments:
            wrapped: The wrapped function.
            factory: An optional factory for creating `self` instances.
        """
        self._wrapped = wrapped
        self._factory = factory
        self._memo = IdMemo[Callable[TParams, TResult]]()

    def __get__(self, owner: TOwner | None, obj_type: type[TOwner]) -> Callable[TParams, TResult]:
        memo = self._memo
        hcurrent = memo.hash(owner, obj_type)
        if hcurrent in memo:
            return self._memo.value

        result: Callable[TParams, TResult]
        wrapped = self._wrapped
        if inspect.isgeneratorfunction(wrapped):
            result = SelfWrapper.sync_generator(wrapped, owner)  # type: ignore[assignment]
        elif inspect.isasyncgenfunction(wrapped):
            result = SelfWrapper.async_generator(wrapped, owner)  # type: ignore[assignment]
        elif asyncio.iscoroutinefunction(wrapped):
            result = SelfWrapper.async_method(wrapped, owner)  # type: ignore[assignment]
        else:
            result = SelfWrapper.sync_method(wrapped, owner)

        replace_self_signature(
            result,
            inspect.Parameter(
                "self",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Annotated[obj_type, Depends(self._factory or obj_type)],
            ),
        )

        return self._memo.store(hcurrent, result)


@overload
def selfdependent(
    *,
    factory: None = None,
) -> Callable[[BoundMethod[TOwner, TParams, TResult]], BoundMethod[TOwner, TParams, TResult]]: ...


@overload
def selfdependent(
    *,
    factory: Dependency[TOwner],
) -> Callable[[BoundMethod[TOwner, TParams, TResult]], BoundMethod[TOwner, TParams, TResult]]: ...


def selfdependent(
    *,
    factory: Dependency[TOwner] | None = None,
) -> Callable[[BoundMethod[TOwner, TParams, TResult]], BoundMethod[TOwner, TParams, TResult]]:
    """
    Decorator that converts an instance method into a FastAPI dependency using a `SelfDependent` descriptor.

    Arguments:
        factory: An optional factory (and FastAPI dependency) that can be wrapped in `Depends()`
            and that produces the `self` argument for the wrapped method.
    """

    def decorator(
        func: Callable[Concatenate[TOwner, TParams], TResult],
    ) -> SelfDependent[TOwner, TParams, TResult]:
        return SelfDependent[TOwner, TParams, TResult](wrapped=func, factory=factory)

    # This type ignore is necessary for mypy in case the decorated method is overridden in a subclass.
    return decorator  # type: ignore



from typing import Any, Generic

from .typing import TResult as TMemo


class IdMemo(Generic[TMemo]):
    """Simple memo class that uses `id()` for hash calculation."""

    __slots__ = ("_hash", "_value")

    def __init__(self) -> None:
        """Initialization."""
        self._hash: int | None = None
        self._value: TMemo | None = None

    def __contains__(self, key: int) -> bool:
        return self._hash == key

    @property
    def value(self) -> TMemo:
        """The current value in the memo."""
        if self._value is None:
            raise KeyError("Memo value accessed before first use.")

        return self._value

    def store(self, key: int, value: TMemo) -> TMemo:
        """
        Stores the given value in the memo.

        Arguments:
            key: The calculated key (hash) for the given `value`.
            value: The value to store in the memo.

        Returns:
            The received `value`.
        """
        self._hash = key
        self._value = value
        return value

    def hash(self, *items: Any) -> int:
        """Calculates the hash of the positional arguments using the `id()` function."""
        return hash(tuple(id(i) for i in items))



from collections.abc import AsyncGenerator, Callable, Generator
from typing import Any, Concatenate, Coroutine, ParamSpec, Protocol, TypeVar

Tcov = TypeVar("Tcov", covariant=True)

TOwner = TypeVar("TOwner")
TResult = TypeVar("TResult")
TParams = ParamSpec("TParams")

BoundMethod = Callable[Concatenate[TOwner, TParams], TResult]


class Dependency(Protocol[Tcov]):
    """Generic FastAPI dependency function protocol."""

    def __call__(
        self, *args: Any, **kwargs: Any
    ) -> Tcov | Coroutine[Any, Any, Tcov] | Generator[Tcov, Any, Any] | AsyncGenerator[Tcov, Any]: ...
